/**
 * @file multiKP.cpp
 * @author Luigi De Giovanni
 * @brief multiKP example implementation
 */

#include "multiKP.hpp"
#include <sstream>
#include "scip_exception.hpp"

using namespace std;
using namespace scipMeMOC;

/* constructor */
scipMeMOC::MultiKPSolver::MultiKPSolver(char * fname)
   : _scip(0), _consCap()
{
  cout << "READING... " << endl;
   // initialize instance data
   readData(fname);
   
   _scip = NULL;
   _numCover = 0;
   _flagInteger = true;
   updateScipModel();
}

void scipMeMOC::MultiKPSolver::updateScipModel(void)
{
  if (_scip != NULL) {
    cleanScip();
  }
  
   //cout << "INIT SCIP... " << endl;
   // initialize scip
   SCIP_CALL_EXC( SCIPcreate(& _scip) );
   // load default plugins like separators, heuristics, etc.
   SCIP_CALL_EXC( SCIPincludeDefaultPlugins(_scip) );
   // disable scip output to stdout
   SCIP_CALL_EXC( SCIPsetMessagehdlr(NULL) );
   // disable presolving
   SCIP_CALL_EXC( SCIPsetIntParam(_scip, "presolving/maxrounds", 0) );
   SCIP_CALL_EXC( SCIPsetIntParam(_scip, "presolving/probing/maxruns", 0) );
   SCIP_CALL_EXC( SCIPsetIntParam(_scip, "propagating/maxroundsroot", 0) );
   SCIP_CALL_EXC( SCIPsetIntParam(_scip, "propagating/maxrounds", 0) );
   SCIP_CALL_EXC( SCIPsetIntParam(_scip, "separating/maxroundsroot", 0) );
   SCIP_CALL_EXC( SCIPsetIntParam(_scip, "separating/maxrounds", 0) );
   SCIP_CALL_EXC( SCIPsetIntParam(_scip, "separating/maxruns", 0) );

   // create an empty problem
   SCIP_CALL_EXC( SCIPcreateProb(_scip, "multiKP", NULL, NULL, NULL, NULL, NULL, NULL) );
   
   // set the objective sense to maximize, default is minimize
   SCIP_CALL_EXC( SCIPsetObjsense(_scip, SCIP_OBJSENSE_MAXIMIZE) );

   // create a variable for every item
   ostringstream namebuf;
   for(int j = 0; j < _numItem; ++j)
   {
       SCIP_VAR * var;
       namebuf.str("");
       namebuf << "x#" << j;
       
       // create the SCIP_VAR object
       SCIP_Vartype varType;
       varType = (_flagInteger ? SCIP_VARTYPE_BINARY : SCIP_VARTYPE_CONTINUOUS);
       SCIP_CALL_EXC( SCIPcreateVar(_scip, & var, namebuf.str().c_str(), 0.0, 1.0, _profit[j], varType, TRUE, FALSE, NULL, NULL, NULL, NULL) );

       // add the SCIP_VAR object to the scip problem
       SCIP_CALL_EXC( SCIPaddVar(_scip, var) );
       
       // storing the SCIP_VAR pointer for later access
       _xload.push_back(var);
   }

   // create constraints
   for (int k = 0; k < _numDim; ++k)
   {
      SCIP_CONS * cons;
      namebuf.str("");
      namebuf<<"dim_"<<k;

      // create SCIP_CONS object
      SCIP_CALL_EXC( SCIPcreateConsLinear(_scip, & cons, namebuf.str().c_str(), 0, NULL, NULL, -SCIPinfinity(_scip), double(_capacity[k]), TRUE, TRUE, TRUE, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE) );

      // add the vars belonging to field in this row to the constraint
      for (int j = 0; j < _numItem; ++j)
         SCIP_CALL_EXC( SCIPaddCoefLinear(_scip, cons, _xload[j], _weight[k][j]) );

      // add the constraint to scip
      SCIP_CALL_EXC( SCIPaddCons(_scip, cons) );

      // store the constraint for later on
      _consCap.push_back(cons);
   }
   
   for (int k = 0 ; k < _numCover ; k++ ) {
     SCIP_CONS * cons;
     namebuf.str("");
     namebuf<<"cov_"<<k;

      // create SCIP_CONS object
     SCIP_CALL_EXC( SCIPcreateConsLinear(_scip, & cons, namebuf.str().c_str(), 0, NULL, NULL, -SCIPinfinity(_scip), _coverCoeff[k][_numItem], TRUE, TRUE, TRUE, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE) );

     //cout << "*** " << namebuf.str() << " *** " << _coverCoeff[k][_numItem] << endl;
     
      // add the vars belonging to field in this row to the constraint
     for (int j = 0; j < _numItem; ++j)
       SCIP_CALL_EXC( SCIPaddCoefLinear(_scip, cons, _xload[j], _coverCoeff[k][j]) );

      // add the constraint to scip
     SCIP_CALL_EXC( SCIPaddCons(_scip, cons) );

      // store the constraint for later on
     _consCap.push_back(cons);
   }

}


SCIP_Real scipMeMOC::MultiKPSolver::getSolution ( std::vector<SCIP_Real> & x  )
{
  SCIP_SOL * sol = SCIPgetBestSol(_scip);
  
  if (sol == NULL) {
    cout << "no solution found" << endl;
    for ( int j = 0 ; j < _numItem; j++ ) {
      x[j] = -1.0;
    }
    return -1.0;
  }
  
  for ( int j = 0 ; j < _numItem; j++ ) {
    x[j] = SCIPgetSolVal(_scip, sol, _xload[j]);
  }
  return SCIPgetSolOrigObj(_scip, sol);
}


void scipMeMOC::MultiKPSolver::addValidCoverInequality ( std::vector<double> & newCover )
{
//cout << "adding cover "<< _numCover << endl;
  int cardinality = 0;
  for (int j = 0; j < _numItem ; j++) {
    cardinality += newCover[j];
  }
  newCover[_numItem] = cardinality - 1;
  _coverCoeff.push_back(newCover);
  
  _numCover++;
  
  updateScipModel();
}



/* display the solution */
void scipMeMOC::MultiKPSolver::display(std::ostream& out , SCIP_SOL * sol )
{
   // get the best found solution from scip
   if (sol == NULL ) sol = SCIPgetBestSol(_scip);
   // when SCIP did not succeed then sol is NULL
   if (sol == NULL)
   {
      out << "no solution found" << endl;
      return;
   }
   
   out << "solution value: " << SCIPgetSolOrigObj(_scip, sol) << endl;

   out << "loaded items (.# index (name), dim , profit):" << endl;
   int n = 0;
   for (int j = 0; j < _numItem; ++j) {
       if ( (SCIPgetSolVal(_scip, sol, _xload[j])) > 0.5 ) {
         out << ++n << "#\t" << j << " (" << _xload[j]->name << " * " << SCIPgetSolVal(_scip, sol, _xload[j]) <<"),\t";
           for ( int k=0; k < _numDim-1; k++ ){
               out << _weight[k][j] << " x "; 
           }
           out << _weight[_numDim-1][j] << " ,\t" << _profit[j] << endl;
       }
   }
   out << endl;

}


/* clean SCIP environment*/
void scipMeMOC::MultiKPSolver::cleanScip(void)
{
//  cout << "CLEANING... " << endl;
  if (_scip == NULL ) {
    return;
  }
  
  // since the SCIPcreateVar captures all variables, we have to release them now
  for (int j = 0; j < _numItem; ++j)
  {
    SCIP_CALL_EXC( SCIPreleaseVar(_scip, & _xload[j]) );
  }
  _xload.clear();

   // the same for the constraints
  for (vector<SCIP_CONS *>::size_type k = 0; k < _consCap.size(); ++k) {
    SCIP_CALL_EXC( SCIPreleaseCons(_scip, &_consCap[k]) );
  }
  _consCap.clear();

  // after releasing all vars and cons we can free the scip problem
  // remember this has allways to be the last call to scip
  SCIP_CALL_EXC( SCIPfree( & _scip) );
}

/* destructor */
scipMeMOC::MultiKPSolver::~MultiKPSolver(void)
{
cout << "DESTROYING... " << endl;
  if (_scip != NULL ) {
    cleanScip();
  }

  _capacity.clear();
  _profit.clear();
  _weight.clear();
  _coverCoeff.clear();
}





/* solve the multiKP problem */
void scipMeMOC::MultiKPSolver::solve(void)
{
//cout << "SOLVING... " << _flagInteger; cout.flush();
   // this tells scip to start the solution process
   SCIP_CALL_EXC( SCIPsolve(_scip) );
//cout << "done!" << _flagInteger << endl;
}
/* relax integrality */
void scipMeMOC::MultiKPSolver::relaxIntegrality ( bool relax )
{
//  cout << "RELAXING INTEGRALITY... " << relax << endl;
  //SCIP_CALL_EXC( SCIPsetLongintParam(_scip, "limits/nodes", (relax ? 0 : -1)) );
  if ( _flagInteger != relax )
    return;
//  cout << "change!!" << endl;
  _flagInteger = !relax;
  updateScipModel();
}


/* verify integrality of the current solution */
bool scipMeMOC::MultiKPSolver::verifyIntegrality ( void )
{
//cout << "VERIFYING INTEGRALITY... " << endl;
  SCIP_SOL * sol = SCIPgetBestSol(_scip);
  // when SCIP did not succeed then sol is NULL
  if (sol == NULL) {
    cout << "no solution found" << endl;
    return false;
  }
  bool isInteger = true;
  for (int j = 0; j < _numItem; ++j) {
//if (SCIPisFeasGT(_scip,SCIPgetSolVal(_scip, sol, _xload[j]),0)) cout << j << ") " << SCIPgetSolVal(_scip, sol, _xload[j]) << endl;
    if ( ! SCIPisFeasIntegral(_scip,SCIPgetSolVal(_scip, sol, _xload[j])) ) {
//cout << "\t*";
      isInteger = false;
     }
//if (SCIPisFeasGT(_scip,SCIPgetSolVal(_scip, sol, _xload[j]),0)) cout << endl;
   }
   return isInteger;
}

SCIP_Real scipMeMOC::MultiKPSolver::getSolVal ( void ) {
  SCIP_SOL * sol = SCIPgetBestSol(_scip);
  if (sol == NULL) {
    cout << "no solution found" << endl;
    return -1.0;
  }
  return SCIPgetSolOrigObj(_scip, sol);
}
      
void scipMeMOC::MultiKPSolver::stats ( void )
{ 
  cout << " ... numNodes: " << SCIPgetNNodes(_scip) << " numRuns: " << SCIPgetNRuns(_scip) << endl;
}
    
    
    
int scipMeMOC::MultiKPSolver::readData(char * fname) {
   int k, j, intBuf;
   FILE * fpdata = fopen(fname,"r");//   int n = abs(atoi(argv[1]));
   if ( fpdata == NULL ) {
       cout << "ERROR reading " << fname << endl;
       exit(-1);
   }
   fscanf(fpdata,"numDim:%d\n",&_numDim);
   fscanf(fpdata,"numItem:%d\n",&_numItem);   
   
   _capacity.reserve(_numDim);
   fscanf(fpdata,"Capacities:\n");
   for (k = 0; k<_numDim; k++) {
       fscanf(fpdata,"%d\n",&intBuf);
       _capacity.push_back(intBuf);
   }
   _profit.reserve(_numItem);
   fscanf(fpdata,"Profits:\n");
   for (j = 0; j<_numItem; j++) {
       fscanf(fpdata,"%d\n",&intBuf);
       _profit.push_back(intBuf);
   }   
   
   fscanf(fpdata,"Weights:\n");
   _weight.reserve(_numDim);
   for (k = 0; k<_numDim; k++) {  
       _weight.push_back(vector<double>());
       for (j = 0; j<_numItem; j++) {
           fscanf(fpdata,"%d",&intBuf);
           _weight[k].push_back(intBuf);
       }
   }
   fclose(fpdata);
   
   return _numDim * _numItem;
}
