/**@file  multiKP.hpp
 * @brief  multiKP example
 * @author Luigi De Giovanni
 */

#ifndef MULTIKP_H
#define MULTIKP_H

#include <vector>
#include <iostream>

#include <scip/scip.h>
#include <scip/scipdefplugins.h>

namespace scipMeMOC
{
   /**@class MultiKPSolver
    * @brief solver class for the multi-dimensional knapsack problem
    *
    *  this class implements a solver for the multi-dimensional knapsack problem as an mip model, which will be solved using SCIP
    */
   class MultiKPSolver
   {
   private:

      /** @brief pointer to scip structure */
      SCIP * _scip;

      /** @brief number of dimensions  */
      int _numDim;
      /** @brief capacity per dimension  */
      std::vector<int> _capacity;

      /** @brief number of items  */
      int _numItem;
      /** @brief profit per item  */
      std::vector<double> _profit;
      /** @brief weight per dimension per item */
      std::vector<std::vector<double> > _weight;

      /** @brief one binary variable for each item */
      std::vector<SCIP_VAR *> _xload;

      /** @brief constraints for knapsasck capacities */
      std::vector<SCIP_CONS *> _consCap;

      int _numCover;
      /** @brief cover vectors and cardinality (in the last position) */
      std::vector<std::vector<double> > _coverCoeff;

   public:
      /** @brief constructs the (integer) programming model for the multi-KP problem
       *
       * @param[in] fname the name of the file containing the instance
       */
      MultiKPSolver ( char * fname );

      /** @brief destructor */
      ~MultiKPSolver();

      /** @brief solves the multi-KP problem using SCIPsolve */
      void solve(void);

      void relaxIntegrality ( bool relax );

      bool verifyIntegrality ( void );
      
      void addValidCoverInequality ( std::vector<double> & newCover );
          
      /** @brief display the solution */
      void display(std::ostream & out = std::cout , SCIP_SOL * sol = NULL );
      
      int getNumItem ( void ) { return _numItem; }
      int getNumDim  ( void ) { return _numDim;  }
      SCIP_Real getSolution ( std::vector<SCIP_Real> & x );
      std::vector<double> & getWeights ( int i ) {
        return _weight[i];
      }
      double getCapacity ( int i ) {
        return _capacity[i];
      }
      
      SCIP_Real getSolVal ( void );
      void stats ( void );
      
   private:
      int readData ( char * fname );

      void updateScipModel ( void );
      void cleanScip ( void );

      bool _flagInteger;

   };
}
#endif

