/**@file multiKP_main.cpp
 * @brief main file for the multiKP example
 * @author Luigi De Giovanni
 */

#include "multiKP.hpp"
#include "CoverSeparator.hpp"
#include <cstdlib>
#include <iostream>
#include "scip_exception.hpp"

using namespace std;
using namespace scipMeMOC;


/** main function for queens example */
int
main(
     int args,
     char ** argv
     )
{
  cout << "********************************************" << endl;
  cout << "* multiKP solver based on SCIP (+ cover)   *" << endl;
  cout << "*                                          *" << endl;
  cout << "* (c) Luigi De Giovanni (2009)             *" << endl;
  cout << "********************************************" << endl << endl;

  if (args < 3)
  {
    cerr << "call " << argv[0] << " <data file>" << " <maxIter>" << endl;
    exit(1);
  }

  try
  {
    int maxIter = atoi(argv[2]);
    int curIter = 0;

    // initialize the multiKP solver
    MultiKPSolver masterMKP(argv[1]);
    
    //solving to integrality
    cout << "SOLVING TO INTEGRALITY (for checking purposes)..." << endl;
    masterMKP.relaxIntegrality(false);
    masterMKP.solve();
    masterMKP.display();
    masterMKP.stats();
    
    //generating cuts
    cout << "\nGENERATING CUTS FOR THE LP RELAXATION..." << endl;
    std::vector<double> newCover(masterMKP.getNumItem()+1); //+1 because cardinality (rhs) in the last position
    SCIP_Real slaveObj = 2.0;
    bool isFeasible = false;
    bool anyViolatedCover = false;
    masterMKP.relaxIntegrality(true);
    do {
      cout << "\n----------- " << ++curIter << "/" << maxIter << " -------------" << endl;
      //solve current LP relaxation
      masterMKP.solve();
      cout << "solution value: " << masterMKP.getSolVal() << endl;
      if ( masterMKP.verifyIntegrality() ) {
        cout << "The solution of the LP relaxation is INTEGER!!!" << endl;
        isFeasible = true;
        break;
      }
      isFeasible = false;
      anyViolatedCover = false;
      //store solution to current LP relaxation
      std::vector<SCIP_Real> x(masterMKP.getNumItem());
      masterMKP.getSolution(x);
      //check for violated cover inequalities (one for KP dimension)
      for ( int i = 0 ; i < masterMKP.getNumDim() ; i++ ) {
        cout << "\tDIMENSION " << i << ", items #: ";
        //solve the separation problem for current dimension
        CoverSeparator KPslave(x,masterMKP.getWeights(i),masterMKP.getCapacity(i));
        KPslave.solve();
        slaveObj = KPslave.getSolution(newCover);
        //if the violation is under 1, add the violated cover inequality
        if (SCIPisFeasLT(KPslave.getScipPtr(),slaveObj,1)) {
          KPslave.displayItems();
          masterMKP.addValidCoverInequality(newCover);
          anyViolatedCover = true;
        } else {
          cout << "no violated ineq.";
        }
        cout << endl;
      }
    } while ( anyViolatedCover && (! isFeasible) && (maxIter-curIter) );

    if (! isFeasible ) {
      //solve to integrality using the LP relaxation sthrengthened by some cover inequalities
      cout << "\nSOLVING THE STHRENGTHENED FORMULATION TO INTEGRALITY..." << endl;
      masterMKP.relaxIntegrality(false);
      masterMKP.solve();
    }
    
    masterMKP.display();
    masterMKP.stats();
    
  } catch(SCIPException& exc)
  {
    cerr << exc.what() << endl;
    exit(exc.getRetcode());
  }
  return EXIT_SUCCESS;
}



//SCIPcheckSol
//separating/maxrounds
//limits/nodes = -1
